'' =================================================================================================
''
''   File....... jm_parser.spin
''   Purpose.... Command input and parsing
''               * token separaters are space, tab, and comma
''   Author..... Jon McPhalen
''               Copyright (c) 2023-2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started.... 19 SEP 2023
''   Updated.... 02 AUG 2024 { 2.2.0 }
''
'' =================================================================================================

{
  General Use (after start())
  - call reset()
  - enter command string with enqueue() or enqueue_str()
  - call parse() (will return # tokens in string)
  - tcount() will retrn # tokens after last parse
  - use token() to return pointer to specific token (string)
  - call token_index() to compare token to list of known words
    * converts token string to index (for case)
    * returns 0..N, or -1 if not found in list
}


con

  VERSION  = 2_2_0

  BUF_SIZE = 64                                                 ' was 128 in 2.1.0
  MAX_TKNS = 10

  #0, NONE, LOWER, UPPER


var

  byte  buffer[BUF_SIZE]                                        ' command buffer
  byte  bidx                                                    ' current entry position in buffer

  long  p_token[MAX_TKNS]                                       ' pointers to tokens in buffer
  byte  ntokens                                                 ' # tokens in buffer (after parsing)

  byte  tchar                                                   ' terminating character (usually CR)
  byte  alphacase                                               ' 0 = none, 1 = lower, 2 = upper


pub null

'' This is not an application


pub start(tc, uc)

'' Prep parser
'' -- tc is input terminating character (usually CR)
'' -- uc is use of case (0 = no change, 1 = lower, 2 = upper)
''    * uc makes token case-sensitive
'' -- p_sep is pointer to string of separator chracters

  tchar := tc                                                   ' get terminating character

  if (uc == LOWER) or (uc == UPPER)                             ' get alpha case option
    alphacase := uc
  else
    alphacase := NONE

  reset


pub reset

'' Reset input buffer and tokens

  bytefill(@buffer, 0, BUF_SIZE)                                ' clear input buffer
  bidx := 0

  longfill(@p_token, 0, MAX_TKNS)                               ' clear tokens
  ntokens := 0


pub enqueue(c) : done | len

'' Add c to input buffer
'' -- returns true when terminating char encountered

  case c
    0, tchar :                                                  ' terminating char?
      buffer[bidx] := 0                                         '  end buffer
      return true

    8 :                                                         ' backspace?
      if (bidx > 0)                                             '  if something in buffer
        buffer[--bidx] := 0                                     '  erase last

    other :
      if (bidx < constant(BUF_SIZE-2))                          ' room in buffer?
        if (alphacase == LOWER)                                 '  check case settings
          c := lcase(c)
        elseif (alphacase == UPPER)
          c := ucase(c)
        buffer[bidx++] := c                                     '  add new character


pub enqueue_str(p_str) : done

'' Add p_str to input buffer
'' -- returns true if terminating char in p_str

  repeat strsize(p_str)
    if enqueue(byte[p_str++])
      return true


pub lcase(c) : lc

'' Convert c to lowercase

  if (c => "A") and (c =< "Z")
    return c + 32
  else
    return c


pub ucase(c) : uc

'' Convert c to uppercase

  if (c => "a") and (c =< "z")
    return c - 32
  else
    return c


pub buf_len : len                                               ' renamed in 2.2.0

'' Returns length of raw input buffer
'' -- will change after parse() if tokens in buffer

  return bidx


pub buf_pntr : p_str                                            ' renamed in 2.2.0

'' Returns pointer to buffer
'' -- after parse() buffer will end after first token

  return @buffer


pub parse : count | p_str, len

'' Parse buffer into separate tokens
'' -- returns token count, 0..MAX_TKNS

  longfill(@p_token, 0, MAX_TKNS)                               ' clear old tokens
  ntokens := 0

  if (strsize(@buffer) == 0)                                    ' abort if nothing to parse
    bidx := 0
    return

  p_str := @buffer                                              ' point to start of input

  repeat MAX_TKNS                                               ' loop through possible tokens
    len := next_token(p_str, @p_str)                            ' get length of next
    if (len > 0)                                                ' good?
      p_token[ntokens++] := p_str                               '  yes, save start of token
      p_str += len+1                                            '  point past end of this token
    else
      quit

  bidx := strsize(@buffer)                                      ' end of buffer is end of 1st token

  return ntokens


pri next_token(p_src, p_head) : len | c                         ' rewritten 01 AUG 2024

'' Extract next token from p_src
'' -- returns length of token if present
'' -- p_head is pointer to start of token address

  repeat                                                        ' clear leading spaces
    c := byte[p_src]                                            ' get byte from string
    if (c == 0)                                                 ' at end?
      return 0                                                  '  abort
    if (is_spacer(c) == false)                                  ' not a spacer
      quit                                                      '  process string
    else
      p_src++                                                   '  skip space char

  long[p_head] := p_src                                         ' point to start of token

  repeat
    c := byte[p_src]
    if (c == 0)                                                 ' end of buffer?
      quit                                                      '  exit
    if (is_spacer(c) == true)
      byte[p_src] := 0                                          ' terminate token in buffer
      quit
    else
      p_src++                                                   ' next char
      len++                                                     ' update length


pri is_spacer(c) : result                                       ' 01 AUG 24 - changed to fixed list

  if (c == " ") or (c == 9) or (c == ",")
    return true


pub tcount : result

'' Returns count of tokens after last parse

  return ntokens


pub token(i) : p_str

'' Returns pointer to token[i]
'' -- valid token pointer will be >0
'' -- 0 if i out of range or token not defined

  if (i => 0) and (i < MAX_TKNS)
    return p_token[i]


pub token_index(p_str, p_list) : idx | len

'' Searches for p_str in p_list of known tokens (see below)
'' -- case sensitive!
'' -- returns 0..N if found
'' -- returns -1 if p_str not in p_list

  repeat
    len := strsize(p_list)                                      ' more in list?
    if (len > 0)
      if (strcomp(p_str, p_list))                               ' check current list item
        return idx                                              ' if match, return index
      else
        p_list += len + 1                                       ' no match, skip to next
        ++idx                                                   ' and bump index
    else
      return -1                                                 ' not found

{
  Example command list
  -- put into DAT section of application
  -- command strings separated by single null (0)
     * put one additional null (0) at end of list
  -- reference this list with @Commands
     * labels allow individual string access (e.g. strcomp())

  Commands      byte
  Cmd0          byte    "cmd0",    0                            ' returns  0
  Cmd1          byte    "cmd1",    0                            ' returns  1
  Cmd2          byte    "cmd2",    0                            ' returns  2
  CmdHelp       byte    "help",    0                            ' returns  3
  CmdVersion    byte    "version", 0                            ' returns  4
  CmdInvalid    byte    0                                       ' returns -1
}

con { numeric support }

pub atoi(p_str, p_err) : result

'' Return value from p_str
'' -- detects decimal, hex ($), binary (%)
''    * first char must be +, -, 0..9, $, %
'' -- p_err is pointer to error flag (long)
'' -- expects trimmed string

  case byte[p_str]
    "0".."9" : result := str2val(p_str,   10, p_err)            ' decimal
    "+", "-" : result := str2val(p_str,   10, p_err)            ' decimal
    "$"      : result := str2val(++p_str, 16, p_err)            ' hexadecimal
    "%"      : result := str2val(++p_str,  2, p_err)            ' binary
    other    : long[p_err] := true                              ' illegal 1st char


pri str2val(p_str, base, p_err) : result | sign, c

'' Convert string to value
'' -- p_str is pointer to string
'' -- base is radix of value
'' -- p_err is pointer to error flag
''    * for detection of bad characters in string

  sign := 1

  if (base == 10)                                               ' if decimal
    c := byte[p_str]                                            '  deal with sign
    if (c == "+")
      p_str++
    if (c == "-")
      p_str++
      sign := -1

  repeat
    c := byte[p_str++]                                          ' get a character
    if (c == 0)                                                 ' if end of string
      long[p_err] := false                                      '  clear error flag
      quit                                                      '  drop out of loop

    if (c == "_")                                               ' skip _ spacer
      next

    c := -15 + --c & %11011111 + 39*(c > 56)                    ' Chip Gracey voodoo?
    if (c => 0) and (c < base)                                  ' valid char for base?
      result := result * base + c                               '  update result
    else
      long[p_err] := true                                       ' bad char, mark error
      return 0                                                  ' abort with 0

  if (base == 10)                                               ' if decimal
    result *= sign                                              '  apply sign


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}