'' =================================================================================================
''
''   File....... jm_parser_demo.spin
''   Purpose....
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2023-2024 Jon McPhalen
''               -- see below for terms of use
''   E-mail..... jon.mcphalen@gmail.com
''   Started....
''   Updated.... 02 AUG 2024
''
'' =================================================================================================


con

  VERSION  = 1_2_1                                              ' 1.2.1


con { timing }

  _xinfreq = 5_000_000                                          ' use 5MHz crystal
  _clkmode = xtal1 + pll16x                                     ' x16 = 80MHz

  CLK_FREQ = (_clkmode >> 6) * _xinfreq                         ' system freq as a constant
  MS_001   = CLK_FREQ / 1_000                                   ' ticks in 1ms
  US_001   = CLK_FREQ / 1_000_000                               ' ticks in 1us


con { terminal }

  BR_TERM  = 115_200                                            ' terminal baud rate


con { fixed io pins }

  PGM_RX   = 31  { I }                                          ' serial / programming
  PGM_TX   = 30  { O }

  EE_SDA   = 29  { I/O }                                        ' i2c / eeprom
  EE_SCL   = 28  { I/O }


con { app io pins }


con

  #true,  ON, OFF
  #false, NO, YES


obj

' main                                                          ' * master Spin cog
  time : "jm_time_80"                                           '   timing and delays (80MHz)
  prng : "jm_prng"                                              '   pseudo-random number generation
  io   : "jm_io"                                                '   simple io
  term : "jm_fullduplexserial"                                  ' * serial IO for terminal
  cli  : "jm_parser"                                            '   token parsing

' * uses cog when loaded


var { globals }

  byte  ntokens                                                 ' tokens in last command


pub main | len, idx

  setup
  wait_for_terminal(true, 250)

  term.fstr0(string("Command line parsing demo\r"))
  term.fstr0(string("-- enter 'help' for commands (case sensitive)\r"))

  repeat
    term.str(string(13, "P1> "))                                ' prompt user
    len := get_command                                          ' get input
    if (len > 0)                                                ' if command entered
      ntokens := cli.parse                                      ' parse and count tokens
      if (ntokens)                                              ' if !0 tokens
        idx := cli.token_index(cli.token(0), @Commands)         ' lookup index of first
        run_command(idx)                                        ' run command code


pub get_command : len | c

'' Enter command from terminal window

  cli.reset                                                     ' prep cli for new entry

  repeat
    c := term.rx                                                ' get char from serial
    if (c == 9)                                                 ' convert tab to space
      c := " "
    if (c == 8)                                                 ' fix backspace
      term.str(string(8, " ", 8))                               '  for Spin Tools terminal
    else
      term.tx(c)                                                ' echo to terminal
    if cli.enqueue(c)                                           ' and to cli queue
      quit                                                      '  if terminated, we're done

  return cli.buf_len                                            ' report length of input


pub run_command(idx)

'' Run code handler

  case idx
    0     : show_help
    1     : show_app_version
    2     : show_parser_version
    3     : term.tx(term#CLS)
    4     : do_cmd0
    5     : do_cmd1
    6     : do_cmd2
    7     : do_cmd3
    8     : do_add
    9     : do_sum
    other : bad_command


pub show_help

' HELP ignores any parameters

  term.str(@s_Help)


pub show_app_version

' VERSION ignores any parameters

  term.fstr3(string("Version %d.%d.%d\r"), {
  }                  VERSION/100,          {
  }                  VERSION//100/10,      {
  }                  VERSION//10)


pub show_parser_version

' PARSER ignores any parameters

  term.fstr3(string("Parser %d.%d.%d\r"), {
  }                  cli#VERSION/100,     {
  }                  cli#VERSION//100/10, {
  }                  cli#VERSION//10)


pub do_cmd0

' CMD0 ignores any parameters

  term.fstr0(string("Running: cmd0\r"))


pub do_cmd1

  if (ntokens <> 2)
    bad_syntax(string("Usage: cmd1 param"))
    return

  term.fstr1(string("Running: cmd1, parameter = %s\r"), cli.token(1))


pub do_cmd2

  if (ntokens <> 3)
    bad_syntax(string("Usage: cmd2 param param"))
    return

  term.fstr2(string("Running: cmd2, parameters = %s, %s\r"), cli.token(1), cli.token(2))


pub do_cmd3

  if (ntokens == 2)
    term.fstr1(string("Running: cmd3, parameter = %s\r"), cli.token(1))
    return

  if (ntokens == 3)
    term.fstr2(string("Running: cmd3, parameters = %s, %s\r"), cli.token(1), cli.token(2))
    return

  if (ntokens == 4)
    term.fstr3(string("Running: cmd3, parameters = %s, %s, %s\r"), {
    }                  cli.token(1), cli.token(2), cli.token(3))
    return

  bad_syntax(string("Usage: cmd3 param [param] [param]"))


pub do_add | v1, v2, t, e

  if (ntokens <> 3)
    bad_syntax(string("Usage: add value value"))
    return

  repeat t from 1 to 2
    v1[t-1] := cli.atoi(cli.token(t), @e)
    if (e)
      term.fstr1(string("Error: %s is not a number\r"), cli.token(t))
      return

  term.fstr3(string("%d + %d = %d\r"), v1, v2, v1+v2)


pub do_sum | s, i, e

  if (ntokens < 3) or (ntokens > 6)
    bad_syntax(string("Usage: sum value value [value] [value] [value]"))
    return

  s := 0
  repeat i from 1 to ntokens-1
    s += cli.atoi(cli.token(i), @e)
    if (e)
      term.fstr2(string("Error: %s is not a number(e = %d)\r"), cli.token(i), e)
      return

  term.fstr1(string("Sum is %d\r"), s)


pub bad_command

  term.fstr1(string("Invalid command (%s)\r"), cli.token(0))


pub bad_syntax(p_str)

  term.fstr0(string("Bad syntax.\r"))
  if (p_str > 0)
    term.fstr1(string("%s\r"), p_str)


dat { help text }

  s_Help        byte    "help       show this help", 13
                byte    "version    show program version", 13
                byte    "parser     show parser version", 13
                byte    "cls        clear terminal", 13
                byte    "cmd0       command with 0 parameters", 13
                byte    "cmd1       command with 1 parameter", 13
                byte    "cmd2       command with 2 parameters", 13
                byte    "cmd3       command with 1, 2, or 3 parameters", 13
                byte    "add        add two integers", 13
                byte    "sum        sum two to five integers", 13
                byte    0


dat { valid commands }

' Caution: Only one null (0) allowed between command words

  Commands      byte    '
  {00}          byte    "help",    0                            ' 0   show commands list
  {01}          byte    "version", 0                            ' 0   show program version #
  {02}          byte    "parser",  0                            ' 0   show parser version #
  {03}          byte    "cls",     0                            ' 0   clear terminal
  {04}          byte    "cmd0",    0                            ' 0   command with 0 args
  {05}          byte    "cmd1",    0                            ' 1   command with 1 arg
  {06}          byte    "cmd2",    0                            ' 2   command with 2 args
  {07}          byte    "cmd3",    0                            ' 1-3 command with 1, 2, or 3 args
  {08}          byte    "add",     0                            ' 2   add two decimal numbers
  {09}          byte    "sum",     0                            ' 2-5 sum two to five numbers
  {end}         byte    0


con { P1 setup / configuration }

pub setup

'' Configure IO and objects for application

  term.tstart(BR_TERM)                                          ' start terminal io

  cli.start(13, cli#NONE)                                       ' setup parser; space or tab separator


pub wait_for_terminal(clear, delay)

'' Wait for terminal to be open and key pressed

  term.rxflush
  term.rx
  if (clear)
    term.tx(term#CLS)
    time.pause(delay #> 0)


con { license }

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

}}